#ifndef bqtExprHH
#define bqtExprHH

#include <string>
#include <vector>

#include "autoptr"

typedef autoptr<class expr> exprp;

#include "tristate"
#include "tokens.hh"
#include "symtables.hh"

class expr: public ptrable
{
public:
    expr();
    virtual ~expr();
    
    tristate is_num() const;
    tristate is_string() const;
    virtual TypeP get_type() const = 0;
    virtual bool is_const() const { return false; }

    virtual exprp optimized()
    {
        return this;
    }
};

class expr_string: public expr
{
    std::string value;
public:
    expr_string() { }
    expr_string(const char* v) : value(v) { }
    expr_string(const std::string& v) : value(v) { }

    virtual TypeP get_type() const { return TypeStr; }
    virtual bool is_const() const  { return true; }
    
    const std::string& get() const { return value; }
};

class expr_number: public expr
{
    TypeP type;
    long long int int_val;
    long double float_val;
    // TODO: value here.
public:
    expr_number() : type(TypeInt), int_val(0), float_val(0) { }
    expr_number(long long int i) { InitIntImpl(i); }
    expr_number(long int i)      { InitIntImpl(i); }
    expr_number(int i)           { InitIntImpl(i); }
    expr_number(long long int i, TypeP t) : type(t),int_val(i),float_val() {}
    expr_number(long int i, TypeP t)      : type(t),int_val(i),float_val() {}
    expr_number(int i, TypeP t)           : type(t),int_val(i),float_val() {}
    expr_number(long double f, TypeP t=TypeDbl)   : type(t),int_val(),float_val(f) {}
    expr_number(double f, TypeP t=TypeDbl)        : type(t),int_val(),float_val(f) {}

    virtual TypeP get_type() const { return type; }
    virtual bool is_const() const  { return true; }

    bool is_int() const;
    long double   get_f() const;
    long long int get_i() const;

private:
    void InitIntImpl(long long int i)
    {
        int_val = i;
        type = (i < -32768 || i > 32767) ? TypeLng : TypeInt;
    }
};


class expr_unary: public expr
{
    QBtoken token;
    exprp p;
public:
	expr_unary(QBtoken tok, exprp a) : token(tok), p(a) { }

    virtual TypeP get_type() const;

    virtual exprp optimized();
};

class expr_binary: public expr
{
    QBtoken token; 
    exprp p0, p1;
public:
	expr_binary(QBtoken tok, exprp a, exprp b) : token(tok), p0(a), p1(b) { }

    virtual TypeP get_type() const;
    /* Note: IMP and EQV have been eliminated in the parse phase. */

    virtual exprp optimized();
};

class expr_label: public expr
{
    long double label;
public:
    expr_label() : label(0) { }
    expr_label(long double f) : label(f) { }
    expr_label(long long int i) : label(i) { }

    virtual TypeP get_type() const { return 0; }
    
    long double get_label() const { return label; }
};

class expr_userfun: public expr
{
    std::string name;
    TypeP type;
    std::vector<exprp> params;
public:
    expr_userfun(const std::string& n) : name(n) { }
    expr_userfun(const std::string& n,
                 TypeP t,
                 const std::vector<exprp>& p) : name(n), type(t), params(p) { }
    
    virtual TypeP get_type() const { return type; }
    
    friend class expr_var;
};

class expr_builtin_numeric: public expr
{
    QBtoken token;
    std::vector<exprp> params;
public:
    expr_builtin_numeric(QBtoken tok) : token(tok) { }
    expr_builtin_numeric(QBtoken tok, exprp a) : token(tok)
        { params.push_back(a); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b) : token(tok)
        { params.push_back(a); params.push_back(b); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b, exprp c) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_numeric(QBtoken tok, exprp a, exprp b, exprp c, exprp d) : token(tok)
        { params.push_back(a); params.push_back(b); params.push_back(c); params.push_back(d); }

    virtual TypeP get_type() const;

    virtual exprp optimized();
};

class expr_builtin_string: public expr
{
    QBtoken token;
    std::vector<exprp> params;
public:
    expr_builtin_string(QBtoken tok) : token(tok) { }
    expr_builtin_string(QBtoken tok, exprp a) : token(tok)
      { params.push_back(a); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b) : token(tok)
      { params.push_back(a); params.push_back(b); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b, exprp c) : token(tok)
      { params.push_back(a); params.push_back(b); params.push_back(c); }
    expr_builtin_string(QBtoken tok, exprp a, exprp b, exprp c, exprp d) : token(tok)
      { params.push_back(a); params.push_back(b);
        params.push_back(c); params.push_back(d); }
    
    void AddParam(exprp p) { params.push_back(p); }
    
    virtual TypeP get_type() const { return TypeStr; }

    virtual exprp optimized();
};

class expr_coordinate: public expr
{
    bool step;
    exprp x, y;
public:
    virtual TypeP get_type() const { return 0; }

    virtual bool is_const() const { return x->is_const() && y->is_const(); }
};

/* These can only occur in select-case */
class expr_caseelse: public expr
{
public:
    expr_caseelse() { }

    virtual TypeP get_type() const { return 0; }

    virtual bool is_const() const { return true; }
};
class expr_caseis: public expr
{
    QBtoken token;
    exprp ref;
public:
    expr_caseis(QBtoken t, exprp r) : token(t), ref(r){ }

    virtual TypeP get_type() const { return 0; }

    virtual bool is_const() const { return ref->is_const(); }
};
class expr_caserange: public expr
{
    exprp min, max;
public:
    expr_caserange(exprp mi, exprp ma) : min(mi), max(ma) { }

    virtual TypeP get_type() const { return 0; }

    virtual bool is_const() const { return min->is_const() && max->is_const(); }
};
class expr_caseor: public expr
{
    exprp a,b;
public:
    expr_caseor(exprp aa, exprp bb) : a(aa),b(bb) { }

    virtual TypeP get_type() const { return 0; }

    virtual bool is_const() const { return a->is_const() && b->is_const(); }
};


class expr_var: public expr
{
    std::string name;
    std::vector<exprp> indices;
    autoptr<expr_var> member;            // NULL if none
    TypeP type;
public:
    expr_var(const std::string& n) : name(n) { }
    
    virtual TypeP get_type() const { return type; }
    virtual bool is_const() const { return false; }
    
    void addindex(exprp p) { indices.push_back(p); }
    void addmember(autoptr<expr_var> m) { member = m; }
    
    bool hasmember() const { return member; }
    bool hasindexes() const { return !indices.empty(); }
    unsigned numindexes() const { return indices.size(); }
    
    const std::string& getname() const { return name; }
    const autoptr<expr_var>& getmember() const { return member; }
    
    void set_type(TypeP t) { type=t; }
    
    autoptr<expr_userfun> convert_to_functioncall(TypeP type) const;
};

class expr_returnvalueholder: public expr_var
{
public :
    explicit expr_returnvalueholder(const std::string& n)
        : expr_var(n) { }
};

class expr_lbound: public expr
{
    exprp var;
    exprp num_index;
public:
    expr_lbound(QBtoken, exprp a, exprp b = new expr_number(1))
        : var(a), num_index(b) { }

    virtual TypeP get_type() const { return TypeLng; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};

class expr_ubound: public expr
{
    exprp var;
    exprp num_index;
public:
    expr_ubound(QBtoken, exprp a, exprp b = new expr_number(1))
        : var(a), num_index(b) { }

    virtual TypeP get_type() const { return TypeLng; }
    virtual bool is_const() const
    {
        if(!num_index->is_const()) return false;
        // Return true if "name" is a $STATIC array, false otherwise.
        return false;
    }
};


#endif
